# detect test environment and add path for local testing
# this should be the first import in __init__.py
from lib_detect_testenv import *

if is_testenv_active():  # pragma: no cover
    add_path_to_syspath(__file__)

# put Your imports here
from .lib_shopware6_api_base import *
from .lib_shopware6_api_base_criteria import *

# __init__conf__ needs to be imported after Your imports, otherwise we would create circular import on the cli script,
# which is reading some values from __init__conf__
from . import __init__conf__

__title__ = __init__conf__.title
__version__ = __init__conf__.version
__name__ = __init__conf__.name
__url__ = __init__conf__.url
__author__ = __init__conf__.author
__author_email__ = __init__conf__.author_email
__shell_command__ = __init__conf__.shell_command



# the configuration file for the dockware docker test container.


"""
# config_dockware{{{

for testing we use the dockware docker container,
see : `dockware <https://developer.shopware.com/docs/guides/installation/dockware>`_


on github actions the dockware docker test container is installed as a service and is available
for communication on localhost

You can start the dockware container locally with the command : sudo docker run -d --rm -p 80:80 --name dockware dockware/dev:latest
# config_dockware}}}
"""


# STDLIB
import os

# conf
try:
    from conf_shopware6_api_base_classes import ConfShopware6ApiBase
except ImportError:  # pragma: no cover
    # Imports for Doctest
    from .conf_shopware6_api_base_classes import ConfShopware6ApiBase  # type: ignore # pragma: no cover

conf_shopware6_api_base = ConfShopware6ApiBase()
# the api url, like : 'https://shop.yourdomain.com/api'
conf_shopware6_api_base.shopware_admin_api_url = "http://localhost/api"
# the storefront api url, like : 'https://shop.yourdomain.com/store-api'
conf_shopware6_api_base.shopware_storefront_api_url = "http://localhost/store-api"

# since dockware container does not support https, we need to disable secure transport for oauth2
os.environ["OAUTHLIB_INSECURE_TRANSPORT"] = "1"


"""
Admin API:
for User Credentials Grant Type:
==================================
- with refresh token
- we recommend to only use this grant flow for client applications that should
  perform administrative actions and require a user-based authentication

"""
conf_shopware6_api_base.username = "admin"
conf_shopware6_api_base.password = "shopware"

"""
Admin API:
for Resource Owner Password Grant Type:
=======================================
- no refresh token
- should be used for machine-to-machine communications, such as CLI jobs or automated services
see https://shopware.stoplight.io/docs/admin-api/ZG9jOjEwODA3NjQx-authentication-and-authorisation
setup via Web Administration Interface > settings > system > integration: "access_id" and "access_secret"
or directly via URL : https://shop.yourdomain.com/admin#/sw/integration/index

those credentials will be created by the APi on the dockware container (fixed, hardcoded)
and are only used for testing purposes on the dockware container on github
"""
# the client ID, setup at Web Administration Interface > settings > system > integration > access_id
conf_shopware6_api_base.client_id = "SWIACWJOMUTXV1RMNGJUAKTUAA"  # noqa
# the client secret, setup at Web Administration Interface > settings > system > integration > access_secret
conf_shopware6_api_base.client_secret = "UkhvUG1qdmpuMjFudGJCdG1Xc0xMbEt2ck9CQ2xDTUtXMUZHRUQ"  # noqa

# which grant type to use can be either 'user_credentials'- or 'resource_owner'
conf_shopware6_api_base.grant_type = "user_credentials"

"""
Store API:
sw-access-key set in Administration/Sales Channels/API
this access key will be read out from the dockware container, and the config will be adjusted accordingly
"""
conf_shopware6_api_base.store_api_sw_access_key = ""



# STDLIB
from datetime import datetime
from typing import Any, Dict, List, Optional, Union

# EXT
# config_class{{{
import attrs
from attrs import validators


@attrs.define
class ConfShopware6ApiBase(object):
    # the api url, like : 'https://shop.yourdomain.com/api'
    shopware_admin_api_url: str = ""
    # the storefront api url, like : 'https://shop.yourdomain.com/store-api'
    shopware_storefront_api_url: str = ""

    """
    Admin API:
    for User Credentials Grant Type:
    ==================================
    - with refresh token
    - we recommend to only use this grant flow for client applications that should
      perform administrative actions and require a user-based authentication

    """
    username: str = ""
    password: str = ""

    """
    Admin API:
    for Resource Owner Password Grant Type:
    =======================================
    - no refresh token
    - should be used for machine-to-machine communications, such as CLI jobs or automated services
    see https://shopware.stoplight.io/docs/admin-api/ZG9jOjEwODA3NjQx-authentication-and-authorisation
    setup via Web Administration Interface > settings > system > integration: "access_id" and "access_secret"
    or directly via URL : https://shop.yourdomain.com/admin#/sw/integration/index
    """
    # the client ID, setup at Web Administration Interface > settings > system > integration > access_id
    client_id: str = ""
    # the client secret, setup at Web Administration Interface > settings > system > integration > access_secret
    client_secret: str = ""

    """
    Admin API:
    Grant Type to use:
    ==================
    which grant type to use - can be either 'user_credentials'- or 'resource_owner'
    """
    grant_type: str = ""

    """
    Store API:
    sw-access-key set in Administration/Sales Channels/API
    """
    store_api_sw_access_key: str = ""


# config_class}}}


class ShopwareAPIError(BaseException):
    pass



# CONF

name = "lib_shopware6_api_base"
title = "python3 base API client for shopware6"
version = "v2.0.8"
url = "https://github.com/bitranox/lib_shopware6_api_base"
author = "Robert Nowotny"
author_email = "bitranox@gmail.com"
shell_command = "lib_shopware6_api_base"


def print_info() -> None:
    print(
        """\

Info for lib_shopware6_api_base:

    python3 base API client for shopware6

    Version : v2.0.8
    Url     : https://github.com/bitranox/lib_shopware6_api_base
    Author  : Robert Nowotny
    Email   : bitranox@gmail.com"""
    )



# STDLIB
from datetime import datetime
from typing import Any, Dict, List, Union

# EXT

import attrs
from attrs import validators

# proj
try:
    from lib_shopware6_api_base_helpers import get_pretty_printer
except ImportError:  # pragma: no cover
    # Imports for Doctest
    from .lib_shopware6_api_base_helpers import get_pretty_printer


@attrs.define
class _EqualFilterTypes:
    """
    definition of Filter Types

    """

    equals: str = "equals"
    equalsAny: str = "equalsAny"
    contains: str = "contains"
    range: str = "range"
    not_: str = "not"
    multi: str = "multi"
    prefix: str = "prefix"
    suffix: str = "suffix"


@attrs.define
class _RangeFilterTypes:
    """
    definition of Range Filter Types


    """

    gte: str = "gte"
    lte: str = "lte"
    gt: str = "gt"
    lt: str = "lt"


@attrs.define
class _NotFilterOperators:
    """
    definition of Not Filter Operators


    """

    or_: str = "or"
    and_: str = "and_"


@attrs.define
class _MultiFilterOperators:
    """
    definition of Multi Filter Operators


    """

    or_: str = "or"
    and_: str = "and_"


equal_filter_type = _EqualFilterTypes()
range_filter = _RangeFilterTypes()
not_filter_operator = _NotFilterOperators()
multi_filter_operator = _MultiFilterOperators()


# EqualsFilter{{{
@attrs.define
class EqualsFilter:
    """
    see filter reference : https://developer.shopware.com/docs/resources/references/core-reference/dal-reference/filters-reference
    The Equals filter allows you to check fields for an exact value.
    The following SQL statement is executed in the background: WHERE stock = 10.

    :parameter:
        field: str
        value: Union[str, int]      # probably also bool

    >>> # Setup
    >>> pp = get_pretty_printer()

    >>> # Test
    >>> my_filter = EqualsFilter('stock', 10)
    >>> pp(attrs.asdict(my_filter))
    {'type': 'equals', 'field': 'stock', 'value': 10}

    >>> my_filter = EqualsFilter('stock', None)
    >>> pp(attrs.asdict(my_filter))
    {'type': 'equals', 'field': 'stock', 'value': None}

    """

    # EqualsFilter}}}

    type: str = attrs.field(init=False, default="equals")
    field: str = attrs.field(validator=attrs.validators.instance_of(str))
    value: Union[None, str, int]


# EqualsAnyFilter{{{
@attrs.define
class EqualsAnyFilter:
    """
    see filter reference : https://developer.shopware.com/docs/resources/references/core-reference/dal-reference/filters-reference
    The EqualsAny filter allows you to filter a field where at least one of the defined values matches exactly.
    The following SQL statement is executed in the background:
    WHERE productNumber IN ('3fed029475fa4d4585f3a119886e0eb1', '77d26d011d914c3aa2c197c81241a45b').

    :parameter:
        field: str
        value: List[str]

    >>> # Setup
    >>> pp = get_pretty_printer()

    >>> # Test Keyword param
    >>> my_filter = EqualsAnyFilter(field = 'productNumber', value = ["3fed029475fa4d4585f3a119886e0eb1", "77d26d011d914c3aa2c197c81241a45b"])
    >>> pp(attrs.asdict(my_filter))
    {'type': 'equals',
     'field': 'productNumber',
     'value': ['3fed029475fa4d4585f3a119886e0eb1',
               '77d26d011d914c3aa2c197c81241a45b']}

    >>> # Test positional param
    >>> my_filter = EqualsAnyFilter('productNumber', ["3fed029475fa4d4585f3a119886e0eb1", "77d26d011d914c3aa2c197c81241a45b"])
    >>> pp(attrs.asdict(my_filter))
    {'type': 'equals',
     'field': 'productNumber',
     'value': ['3fed029475fa4d4585f3a119886e0eb1',
               '77d26d011d914c3aa2c197c81241a45b']}

    """

    # EqualsAnyFilter}}}

    type: str = attrs.field(init=False, default="equals")
    field: str = attrs.field(init=True, validator=attrs.validators.instance_of(str))
    value: List[str] = attrs.field(init=True, factory=list)


# ContainsFilter{{{
@attrs.define
class ContainsFilter:
    """
    see filter reference : https://developer.shopware.com/docs/resources/references/core-reference/dal-reference/filters-reference
    The Contains Filter allows you to filter a field to an approximate value, where the passed value must be contained as a full value.
    The following SQL statement is executed in the background: WHERE name LIKE '%Lightweight%'.

    :parameter:
        field: str
        value: List[str]

    >>> # Setup
    >>> pp = get_pretty_printer()

    >>> # Test
    >>> my_filter = ContainsFilter(field = 'productNumber', value = 'Lightweight')
    >>> pp(attrs.asdict(my_filter))
    {'type': 'contains', 'field': 'productNumber', 'value': 'Lightweight'}


    """

    # ContainsFilter}}}
    type: str = attrs.field(init=False, default="contains")
    field: str = attrs.field(init=True, validator=attrs.validators.instance_of(str))
    value: List[str] = attrs.field(init=True, factory=list)


# RangeFilter{{{
@attrs.define
class RangeFilter:
    """
    see filter reference : https://developer.shopware.com/docs/resources/references/core-reference/dal-reference/filters-reference
    The Range filter allows you to filter a field to a value space. This can work with date or numerical values.
    Within the parameter property the following values are possible:
        gte => Greater than equals  (You might pass 'gte' or range_filter.gte)
        lte => Less than equals     (You might pass 'lte' or range_filter.lte)
        gt => Greater than          (You might pass 'gt' or range_filter.gt)
        lt => Less than             (You might pass 'lt' or range_filter.lt)

    :parameter:
        field: str
        parameters: Dict[str, Union[int, datetime]]

    >>> # Setup
    >>> pp = get_pretty_printer()

    >>> # Test (pass range type as string)
    >>> my_filter = RangeFilter(field = 'stock', parameters = {'gte': 20, 'lte': 30})
    >>> pp(attrs.asdict(my_filter))
    {'type': 'range', 'field': 'stock', 'parameters': {'gte': 20, 'lte': 30}}

    >>> # Test (pass range type from 'range_filter' object)
    >>> my_filter = RangeFilter(field = 'stock', parameters = {range_filter.gte: 20, range_filter.lte: 30})
    >>> pp(attrs.asdict(my_filter))
    {'type': 'range', 'field': 'stock', 'parameters': {'gte': 20, 'lte': 30}}

    >>> # Test (wrong range)
    >>> my_filter = RangeFilter(field = 'stock', parameters = {'gte': 20, 'less': 30})
    Traceback (most recent call last):
        ...
    ValueError: "less" is not a valid range

    """

    # RangeFilter}}}
    type: str = attrs.field(init=False, default="range")
    field: str = attrs.field(validator=attrs.validators.instance_of(str))
    parameters: Dict[str, Union[int, datetime]] = attrs.field(validator=validators.instance_of(dict), factory=dict)

    @parameters.validator  # noqa
    def check(self, attribute: attrs.Attribute, parameters: Dict[str, Any]) -> None:  # type: ignore
        for parameter in parameters:
            if parameter not in ["gte", "lte", "gt", "lt"]:
                raise ValueError(f'"{parameter}" is not a valid range')


# NotFilter{{{
@attrs.define
class NotFilter:
    """
    see filter reference : https://developer.shopware.com/docs/resources/references/core-reference/dal-reference/filters-reference
    The Not Filter is a container which allows to negate any kind of filter.
    The operator allows you to define the combination of queries within the NOT filter (OR and AND).
    The following SQL statement is executed in the background: WHERE !(stock = 1 OR availableStock = 1):

    :parameter:
        operator: 'or' | 'and'
        queries: List[Filter]

    >>> # Setup
    >>> pp = get_pretty_printer()

    >>> # Test (pass operator as string)
    >>> my_filter = NotFilter('or', [EqualsFilter('stock', 1), EqualsFilter('availableStock', 10)])
    >>> pp(attrs.asdict(my_filter))
    {'type': 'not',
     'operator': 'or',
     'queries': [{'type': 'equals', 'field': 'stock', 'value': 1},
                 {'type': 'equals', 'field': 'availableStock', 'value': 10}]}

    >>> # Test (pass operator from 'not_filter_operator' object)
    >>> my_filter = NotFilter(not_filter_operator.or_, [EqualsFilter('stock', 1), EqualsFilter('availableStock', 10)])
    >>> pp(attrs.asdict(my_filter))
    {'type': 'not',
     'operator': 'or',
     'queries': [{'type': 'equals', 'field': 'stock', 'value': 1},
                 {'type': 'equals', 'field': 'availableStock', 'value': 10}]}

    >>> # Test unknown operator
    >>> my_filter = NotFilter('duck', [EqualsFilter('stock', 1), EqualsFilter('availableStock', 10)])
    Traceback (most recent call last):
        ...
    ValueError: 'operator' must be in ['and', 'or'] (got 'duck')

    """

    # NotFilter}}}
    type: str = attrs.field(init=False, default="not")
    operator: str = attrs.field(init=True, validator=attrs.validators.in_(["and", "or"]))
    queries: List["FilterType"] = attrs.field(factory=list)


# MultiFilter{{{
@attrs.define
class MultiFilter:
    """
    see filter reference : https://developer.shopware.com/docs/resources/references/core-reference/dal-reference/filters-reference
    The Multi Filter is a container, which allows to set logical links between filters.
    The operator allows you to define the links between the queries within the Multi filter (OR and AND).
    The following SQL statement is executed in the background: WHERE (stock = 1 OR availableStock = 1)

    :parameter:
        operator: 'or' | 'and'
        queries: List[Filter]

    >>> # Setup
    >>> pp = get_pretty_printer()

    >>> # Test (pass operator as string)
    >>> my_filter = MultiFilter('or', [EqualsFilter('stock', 1), EqualsFilter('availableStock', 10)])
    >>> pp(attrs.asdict(my_filter))
    {'type': 'multi',
     'operator': 'or',
     'queries': [{'type': 'equals', 'field': 'stock', 'value': 1},
                 {'type': 'equals', 'field': 'availableStock', 'value': 10}]}

    >>> # Test (pass operator from 'not_filter_operator' object)
    >>> my_filter = MultiFilter(multi_filter_operator.or_, [EqualsFilter('stock', 1), EqualsFilter('availableStock', 10)])
    >>> pp(attrs.asdict(my_filter))
    {'type': 'multi',
     'operator': 'or',
     'queries': [{'type': 'equals', 'field': 'stock', 'value': 1},
                 {'type': 'equals', 'field': 'availableStock', 'value': 10}]}

    >>> # Test unknown operator
    >>> my_filter = MultiFilter('duck', [EqualsFilter('stock', 1), EqualsFilter('availableStock', 10)])
    Traceback (most recent call last):
        ...
    ValueError: 'operator' must be in ['and', 'or'] (got 'duck')

    """

    # MultiFilter}}}

    type: str = attrs.field(init=False, default="multi")
    operator: str = attrs.field(init=True, validator=attrs.validators.in_(["and", "or"]))
    queries: List["FilterType"] = attrs.field(factory=list)


# PrefixFilter{{{
@attrs.define
class PrefixFilter:
    """
    see filter reference : https://developer.shopware.com/docs/resources/references/core-reference/dal-reference/filters-reference
    The Prefix Filter allows you to filter a field to an approximate value, where the passed value must be the start of a full value.
    The following SQL statement is executed in the background: WHERE name LIKE 'Lightweight%'.

    :parameter:
        field: str
        value: str

    >>> # Setup
    >>> pp = get_pretty_printer()

    >>> # Test
    >>> my_filter = PrefixFilter('name', 'Lightweight')
    >>> pp(attrs.asdict(my_filter))
    {'type': 'prefix', 'field': 'name', 'value': 'Lightweight'}

    """

    # PrefixFilter}}}

    type: str = attrs.field(init=False, default="prefix")
    field: str = attrs.field(init=True)
    value: str = attrs.field(init=True)


# SuffixFilter{{{
@attrs.define
class SuffixFilter:
    """
    see filter reference : https://developer.shopware.com/docs/resources/references/core-reference/dal-reference/filters-reference
    The Suffix Filter allows you to filter a field to an approximate value, where the passed value must be the end of a full value.
    The following SQL statement is executed in the background: WHERE name LIKE '%Lightweight'.

    :parameter:
        field: str
        value: str

    >>> # Setup
    >>> pp = get_pretty_printer()

    >>> # Test
    >>> my_filter = SuffixFilter('name', 'Lightweight')
    >>> pp(attrs.asdict(my_filter))
    {'type': 'suffix', 'field': 'name', 'value': 'Lightweight'}

    """

    # SuffixFilter}}}

    type: str = attrs.field(init=False, default="suffix")
    field: str = attrs.field(init=True)
    value: str = attrs.field(init=True)


FilterType = Union[EqualsFilter, EqualsAnyFilter, ContainsFilter, RangeFilter, NotFilter, MultiFilter, PrefixFilter, SuffixFilter]



# STDLIB
import os
from functools import lru_cache
import json

import logging
import sys
import time
from typing import Any, Dict, List, Optional, Union, cast

# EXT
import oauthlib
from oauthlib.oauth2 import BackendApplicationClient, TokenExpiredError
import requests
import requests_oauthlib

# conf
try:
    from conf_shopware6_api_base_classes import *
    from lib_shopware6_api_base_criteria import *
except ImportError:  # pragma: no cover
    # Imports for Doctest
    from .conf_shopware6_api_base_classes import *  # type: ignore  # pragma: no cover
    from .lib_shopware6_api_base_criteria import *


logger = logging.getLogger(__name__)

# payload_type{{{
PayLoad = Union[None, Dict[str, Any], Criteria]
# payload_type}}}


# store_api{{{
class Shopware6StoreFrontClientBase(object):
    def __init__(self, config: Optional[ConfShopware6ApiBase] = None, use_docker_test_container: bool = False) -> None:
        """
        the Shopware6 Storefront Base API

        :param config:  You can pass a configuration object here.
                        If not given and github actions is detected, or use_docker_test_container == True:
                            conf_shopware6_api_docker_testcontainer.py will be loaded automatically
                        If not given and no github actions is detected:
                            conf_shopware6_api_base_rotek.py will be loaded automatically

        :param use_docker_test_container:   if True, and no config is given, the dockware config will be loaded

        >>> # Test to load automatic configuration
        >>> my_storefront_client = Shopware6StoreFrontClientBase()

        >>> # Test pass configuration
        >>> if _is_github_actions():
        ...     my_config = _load_config_for_docker_test_container()
        ...     my_storefront_client = Shopware6StoreFrontClientBase(config=my_config)

        """
        # store_api}}}
        self.use_docker_test_container = use_docker_test_container

        if config is None:
            config = _load_config(use_docker_test_container=use_docker_test_container)

        self.config = config

    # store_api_delete{{{
    def request_delete(self, request_url: str, payload: PayLoad = None) -> Dict[str, Any]:
        """
        make a delete request

        parameters:
            http_method: get, post, put, delete
            request_url: API Url, without the common api prefix
            payload : a dictionary

        :returns
            response_dict: dictionary with the response as dict

        """
        # store_api_delete}}}
        response_dict = self._request_dict(http_method="delete", request_url=request_url, payload=payload)
        return response_dict

    # store_api_get{{{
    def request_get(self, request_url: str, payload: PayLoad = None) -> Dict[str, Any]:
        """
        make a get request

        parameters:
            request_url: API Url, without the common api prefix
            payload : a dictionary

        :returns
            response_dict: dictionary with the response as dict

        >>> # Setup
        >>> my_storefront_client = Shopware6StoreFrontClientBase()

        >>> # test GET a dictionary
        >>> my_response = my_storefront_client.request_get(request_url='context')

        >>> # test GET a List
        >>> my_response = my_storefront_client.request_get(request_url='sitemap')
        Traceback (most recent call last):
            ...
        conf_shopware6_api_base_classes.ShopwareAPIError: received a list instead of a dict - You need to use the method request_get_list

        """
        # store_api_get}}}
        response_dict = self._request_dict(http_method="get", request_url=request_url, payload=payload)
        return response_dict

    # store_api_get_list{{{
    def request_get_list(self, request_url: str, payload: PayLoad = None) -> List[Dict[str, Any]]:
        """
        make a get request, expecting a list of dictionaries as result

        parameters:
            request_url: API Url, without the common api prefix
            payload : a dictionary

        :returns
            List[response_dict]: a list of dictionaries

        >>> # Setup
        >>> my_storefront_client = Shopware6StoreFrontClientBase()

        >>> # test GET a List
        >>> my_response = my_storefront_client.request_get_list(request_url='sitemap')

        >>> # test GET a dictionary
        >>> my_response = my_storefront_client.request_get_list(request_url='context')
        Traceback (most recent call last):
            ...
        conf_shopware6_api_base_classes.ShopwareAPIError: received a dict instead of a list - You need to use the method request_get


        """
        # store_api_get_list}}}
        response_l_dict = self._request_list(http_method="get", request_url=request_url, payload=payload)
        return response_l_dict

    # store_api_patch{{{
    def request_patch(self, request_url: str, payload: PayLoad = None) -> Dict[str, Any]:
        """
        makes a patch request

        parameters:
            request_url: API Url, without the common api prefix
            payload : a dictionary

        :returns
            response_dict: dictionary with the response as dict

        """
        # store_api_patch}}}
        response_dict = self._request_dict(http_method="patch", request_url=request_url, payload=payload)
        return response_dict

    # store_api_post{{{
    def request_post(self, request_url: str, payload: PayLoad = None) -> Dict[str, Any]:
        """
        make a post request

        parameters:
            request_url: API Url, without the common api prefix
            payload : a dictionary

        :returns
            response_dict: dictionary with the response as dict

        >>> # Setup
        >>> my_storefront_client = Shopware6StoreFrontClientBase()

        >>> # test POST without payload
        >>> my_response = my_storefront_client.request_post(request_url='product')
        >>> assert 'elements' in my_response

        >>> # test POST with payload
        >>> # see : https://shopware.stoplight.io/docs/store-api/b3A6ODI2NTY4MQ-fetch-a-list-of-products
        >>> my_payload = Criteria()
        >>> my_payload.filter.append(EqualsFilter(field='active', value='true'))
        >>> my_response = my_storefront_client.request_post(request_url='product', payload=my_payload)
        >>> assert 'elements' in my_response

        """
        # store_api_post}}}

        response_dict = self._request_dict(http_method="post", request_url=request_url, payload=payload)
        return response_dict

    # store_api_put{{{
    def request_put(self, request_url: str, payload: PayLoad = None) -> Dict[str, Any]:
        """
        make a put request

        parameters:
            http_method: get, post, put, delete
            request_url: API Url, without the common api prefix
            payload : a dictionary

        :returns
            response_dict: dictionary with the response as dict

        """
        # store_api_put}}}
        response_dict = self._request_dict(http_method="put", request_url=request_url, payload=payload)
        return response_dict

    def _request_dict(self, http_method: str, request_url: str, payload: PayLoad = None) -> Dict[str, Any]:
        """
        http requests a dictionary. raises ShopwareAPIError if the result is not a dictionary
        :param http_method:
        :param request_url:
        :param payload:
        :return:
        """
        response = self._request(http_method=http_method, request_url=request_url, payload=payload)
        if hasattr(response, "json"):  # pragma: no cover
            response_json = response.json()  # type: ignore
            if isinstance(response_json, list):
                raise ShopwareAPIError(f"received a list instead of a dict - You need to use the method request_{http_method}_list")
            response_dict = dict(response_json)
        else:
            response_dict = dict()  # pragma: no cover
        return response_dict

    def _request_list(self, http_method: str, request_url: str, payload: PayLoad = None) -> List[Dict[str, Any]]:
        """
        http requests a list of dictionaries. raises ShopwareAPIError if the result is not a list
        :param http_method:
        :param request_url:
        :param payload:
        :return:
        """
        response = self._request(http_method=http_method, request_url=request_url, payload=payload)
        if hasattr(response, "json"):  # pragma: no cover
            response_json = response.json()  # type: ignore
            if isinstance(response_json, dict):
                raise ShopwareAPIError(f"received a dict instead of a list - You need to use the method request_{http_method}")
            response_l_dict = list(response_json)
        else:
            response_l_dict = list()  # pragma: no cover
        return response_l_dict

    def _request(self, http_method: str, request_url: str, payload: PayLoad = None) -> Optional[requests.Response]:
        """
        makes a request, using the conf.store_api_sw_access_key for authentication

        parameters:
            http_method: 'get', 'patch', 'post', 'put', 'delete'
            request_url: API Url, without the common api prefix
            payload : a dictionary

        :returns
            response_dict: dictionary with the response as dict


        >>> # Setup
        >>> my_api_client = Shopware6StoreFrontClientBase()

        >>> # test GET (a new cart)
        >>> my_api_client._request(http_method='get', request_url='checkout/cart', payload={'name': 'rn_doctest'})
        <Response [200]>

        >>> # test DELETE (a new cart)
        >>> my_api_client._request(http_method='delete', request_url='checkout/cart')
        <Response [204]>

        >>> # test POST (a single product)
        >>> my_payload=Criteria(filter=[EqualsFilter(field='active', value='true')])
        >>> my_api_client._request(http_method='post', request_url='product', payload=my_payload)
        <Response [200]>

        >>> # test Link does not exist (a single product)
        >>> my_api_client._request(http_method='post', request_url='product/000')
        Traceback (most recent call last):
            ...
        conf_shopware6_api_base_classes.ShopwareAPIError: 400 Client Error: Bad Request for url: ...

        """
        if isinstance(payload, Criteria):
            payload = payload.get_dict()
        formatted_request_url = self._format_storefront_api_url(request_url)
        response: requests.Response = requests.Response()
        headers = self._get_headers()

        if http_method == "get":
            response = requests.request("GET", formatted_request_url, params=payload, headers=headers)
        elif http_method == "patch":
            response = requests.request("PATCH", formatted_request_url, data=json.dumps(payload), headers=headers)
        elif http_method == "post":
            response = requests.request("POST", formatted_request_url, data=json.dumps(payload), headers=headers)
        elif http_method == "put":
            response = requests.request("PUT", formatted_request_url, data=json.dumps(payload), headers=headers)
        elif http_method == "delete":
            response = requests.request("DELETE", formatted_request_url, headers=headers)

        try:
            response.raise_for_status()
        except Exception as exc:
            if hasattr(exc, "response"):  # pragma: no cover
                detailed_error = f" : {exc.response.text}"  # type: ignore
            else:
                detailed_error = ""  # pragma: no cover
            raise ShopwareAPIError(f"{exc}{detailed_error}")
        return response

    @lru_cache(maxsize=None)
    def _get_headers(self) -> Dict[str, str]:
        """
        >>> my_api_client = Shopware6StoreFrontClientBase()
        >>> my_api_client._get_headers()
        {'Content-Type': 'application/json', 'Accept': 'application/json', 'sw-access-key': '...'}
        >>> my_api_client._get_headers.cache_clear()

        """
        headers = {"Content-Type": "application/json", "Accept": "application/json", "sw-access-key": self.config.store_api_sw_access_key}
        return headers

    def _format_storefront_api_url(self, request_url: str) -> str:
        """
        formatted url to make a request

        :parameter
            request_url:                        the request url, for instance "oauth/token"
            self.shopware_storefront_api_url:   the api url, for instance https://your.shop-domain.com/store-api

        :returns
            the formatted url, like  https://your.shop-domain.com/store-api/product

        >>> my_api_client = Shopware6StoreFrontClientBase()
        >>> my_api_client._format_storefront_api_url('test')
        'http.../store-api/test'

        """
        request_url = request_url.lstrip("/")
        return f"{self.config.shopware_storefront_api_url}/{request_url}"


# admin_api{{{
class Shopware6AdminAPIClientBase(object):
    def __init__(self, config: Optional[ConfShopware6ApiBase] = None, use_docker_test_container: bool = False) -> None:
        """
        the Shopware6 Admin Base API

        :param config:  You can pass a configuration object here.
                If not given and github actions is detected, or use_docker_test_container == True:
                    conf_shopware6_api_docker_testcontainer.py will be loaded automatically
                If not given and no github actions is detected:
                    conf_shopware6_api_base_rotek.py will be loaded automatically

        :param use_docker_test_container:   if True, and no config is given, the dockware config will be loaded

        >>> # Setup
        >>> my_api_client = Shopware6AdminAPIClientBase()

        """
        # admin_api}}}

        self.use_docker_test_container = use_docker_test_container
        if config is None:
            config = _load_config(use_docker_test_container=use_docker_test_container)
        self.config = config
        self.token: Dict[str, Any] = dict()
        self.session: requests_oauthlib.OAuth2Session = requests_oauthlib.OAuth2Session()

    # admin_api_get{{{
    def request_get(self, request_url: str, payload: PayLoad = None) -> Dict[str, Any]:
        """
        makes a get request

        parameters:
            request_url: API Url, without the common api prefix
            payload : a dictionary

        :returns
            response_dict: dictionary with the response as dict

        >>> # Setup
        >>> my_api_client = Shopware6AdminAPIClientBase()

        >>> # test resource owner token
        >>> ignore = my_api_client._get_access_token_by_user_credentials()
        >>> my_api_client._get_session()
        >>> ignore = my_api_client.request_get('customer-group')  # noqa

        >>> # test resource owner token refresh
        >>> my_access_token = my_api_client.token['access_token']
        >>> my_api_client.token['expires_in']=-1
        >>> my_api_client.token['expires_at']=time.time()-1
        >>> ignore = my_api_client.request_get('customer-group')
        >>> assert my_api_client.token['access_token'] != my_access_token

        >>> # Test client credentials token
        >>> ignore = my_api_client._get_access_token_by_resource_owner()
        >>> my_api_client._get_session()
        >>> ignore = my_api_client.request_get('customer-group')  # noqa

        >>> # test client credentials token refresh
        >>> my_access_token = my_api_client.token['access_token']
        >>> my_api_client.token['expires_in']=-1
        >>> my_api_client.token['expires_at']=time.time()-1
        >>> ignore = my_api_client.request_get('customer-group')
        >>> assert my_api_client.token['access_token'] != my_access_token

        """
        # admin_api_get}}}
        response_dict = self._make_request(http_method="get", request_url=request_url, payload=payload)
        return response_dict

    # admin_api_get_paginated{{{
    def request_get_paginated(self, request_url: str, payload: PayLoad = None, junk_size: int = 100) -> Dict[str, Any]:
        """
        get the data paginated - metadata 'total' and 'totalCountMode' will be updated
        the paginated request reads those records in junks of junk_size=100 for performance reasons.

        payload "limit" will be respected (meaning we deliver only 'limit' results back)
        payload "page" will be ignored

        parameters:
            request_url: API Url, without the common api prefix
            payload : a dictionary
            limit : the junk size

        :returns
            response_dict: dictionary with the response as dict

        >>> # Setup
        >>> my_api_client = Shopware6AdminAPIClientBase()

        >>> # test read product junk_size=3, limit = 4
        >>> my_payload={'limit': 4}
        >>> my_response_dict = my_api_client.request_get_paginated(request_url='product', payload=my_payload, junk_size=3)
        >>> assert 4 == len(my_response_dict['data'])

        >>> # test read product junk_size=3, no limit
        >>> my_response_dict = my_api_client.request_get_paginated(request_url='product', junk_size=3)
        >>> assert 3 < len(my_response_dict['data'])

        >>> # test read product junk_size=3, limit = 2
        >>> my_payload={'limit': 2}
        >>> my_response_dict = my_api_client.request_get_paginated(request_url='product', payload=my_payload, junk_size=3)
        >>> assert 2 == len(my_response_dict['data'])

        >>> # test read product junk_size=3, limit = 4
        >>> my_payload={'limit': 4}
        >>> my_response_dict = my_api_client.request_get_paginated(request_url='product', payload=my_payload, junk_size=3)
        >>> assert 4 == len(my_response_dict['data'])

        >>> # test read product junk_size=10, limit = None
        >>> my_payload=Criteria()
        >>> my_response_dict = my_api_client.request_get_paginated(request_url='product', payload=my_payload, junk_size=10)
        >>> assert 5 < len(my_response_dict['data'])


        """
        # admin_api_get_paginated}}}
        response_dict = self._request_paginated(http_method="get", request_url=request_url, payload=payload, junk_size=junk_size)
        return response_dict

    # admin_api_patch{{{
    def request_patch(self, request_url: str, payload: PayLoad = None, content_type: str = "json") -> Dict[str, Any]:
        """
        makes a patch request

        parameters:
            request_url: API Url, without the common api prefix
            payload : a dictionary or bytes
            content_type: any valid content type like json, octet-stream, ...

        :returns
            response_dict: dictionary with the response as dict

        """
        # admin_api_patch}}}
        response_dict = self._make_request(http_method="patch", request_url=request_url, payload=payload, content_type=content_type)
        return response_dict

    # admin_api_post{{{
    def request_post(self, request_url: str, payload: PayLoad = None, content_type: str = "json") -> Dict[str, Any]:
        """
        makes a post request

        parameters:
            request_url: API Url, without the common api prefix
            payload : a dictionary or bytes
            content_type: any valid content type like json, octet-stream, ...

        :returns
            response_dict: dictionary with the response as dict

        """
        # admin_api_post}}}
        response_dict = self._make_request(http_method="post", request_url=request_url, payload=payload, content_type=content_type)
        return response_dict

    # admin_api_post_paginated{{{
    def request_post_paginated(self, request_url: str, payload: PayLoad = None, junk_size: int = 100) -> Dict[str, Any]:
        """
        post the data paginated - metadata 'total' and 'totalCountMode' will be updated
        if You expect a big number of records, the paginated request reads those records in junks of junk_size=100 for performance reasons.

        payload "limit" will be respected (meaning we deliver only 'limit' results back)
        payload "page" will be ignored

        parameters:
            request_url: API Url, without the common api prefix
            payload : a dictionary
            junk_size : the junk size

        :returns
            response_dict: dictionary with the response as dict

        >>> # Setup
        >>> my_api_client = Shopware6AdminAPIClientBase()
        >>> my_url = 'search/product'

        >>> # test read product junk_size=10, limit = None
        >>> my_payload=Criteria()
        >>> my_response_dict = my_api_client.request_post_paginated(request_url=my_url, payload=my_payload, junk_size=10)
        >>> assert 5 < len(my_response_dict['data'])

        >>> # test read product junk_size=10, no limit
        >>> my_payload=None
        >>> my_response_dict = my_api_client.request_post_paginated(request_url=my_url, payload=my_payload, junk_size=10)
        >>> assert 10 < len(my_response_dict['data'])

        >>> # test read product junk_size=3, limit = 2
        >>> my_payload={'limit': 2}
        >>> my_response_dict = my_api_client.request_post_paginated(request_url=my_url, payload=my_payload, junk_size=3)
        >>> assert 2 == len(my_response_dict['data'])

        >>> # test read product junk_size=3, limit = 4
        >>> my_payload={'limit': 4}
        >>> my_response_dict = my_api_client.request_post_paginated(request_url=my_url, payload=my_payload, junk_size=3)
        >>> assert 4 == len(my_response_dict['data'])

        """
        # admin_api_post_paginated}}}
        response_dict = self._request_paginated(http_method="post", request_url=request_url, payload=payload, junk_size=junk_size)
        return response_dict

    # admin_api_put{{{
    def request_put(self, request_url: str, payload: PayLoad = None, content_type: str = "json") -> Dict[str, Any]:
        """
        makes a put request

        parameters:
            http_method: get, post, put, delete
            request_url: API Url, without the common api prefix
            payload : a dictionary or bytes
            content_type: any valid content type like json, octet-stream, ...

        :returns
            response_dict: dictionary with the response as dict

        """
        # admin_api_put}}}
        response_dict = self._make_request(http_method="put", request_url=request_url, payload=payload, content_type=content_type)
        return response_dict

    # admin_api_delete{{{
    def request_delete(self, request_url: str, payload: PayLoad = None) -> Dict[str, Any]:
        """
        makes a delete request

        parameters:
            http_method: get, post, put, delete
            request_url: API Url, without the common api prefix
            payload : a dictionary

        :returns
            response_dict: dictionary with the response as dict

        """
        # admin_api_delete}}}
        response_dict = self._make_request(http_method="delete", request_url=request_url, payload=payload)
        return response_dict

    def _request_paginated(self, http_method: str, request_url: str, payload: PayLoad = None, junk_size: int = 100) -> Dict[str, Any]:
        """
        request the data paginated - metadata 'total' and 'totalCountMode' will be updated
        the paginated request reads those records in junks of junk_size=100 for performance reasons.

        payload "limit" will be respected (meaning we deliver only 'limit' results back)
        "page" will be ignored

        parameters:
            http_method:
            request_url: API Url, without the common api prefix
            payload : a dictionary
            junk_size : the junk size

        :returns
            response_dict: dictionary with the response as dict


        :param http_method:
        :param request_url:
        :param payload:
        :param junk_size:
        :return:
        """
        response_dict: Dict[str, Any] = dict()
        response_dict["data"] = list()
        payload_dict = _get_payload_dict(payload)

        # when 'ids' are given, limit does not apply, so we need to set
        # 'limit' and 'junk_size' to len(payload.ids)
        if _is_type_criteria(payload):
            if payload.ids:  # type: ignore
                ids_limit = len(payload.ids)  # type: ignore
                junk_size = ids_limit
                payload_dict["limit"] = ids_limit

        total_limit: Union[None, int]
        records_left: Union[None, int]

        if "limit" in payload_dict:
            total_limit = payload_dict["limit"]
        else:
            total_limit = None

        if total_limit is None:
            payload_dict["limit"] = str(junk_size)
            records_left = 0
        else:
            payload_dict["limit"] = min(total_limit, junk_size)
            records_left = total_limit

        page = 1

        while True:
            payload_dict["page"] = page

            partial_data = self._make_request(http_method=http_method, request_url=request_url, payload=payload_dict)
            if partial_data["data"]:
                response_dict["data"] = response_dict["data"] + partial_data["data"]
                page = page + 1
                if total_limit is not None:
                    records_left = records_left - len(partial_data["data"])
                    if records_left < 1:
                        response_dict["data"] = response_dict["data"][:total_limit]
                        break
            else:
                break
        return response_dict

    def _make_request(self, http_method: str, request_url: str, payload: PayLoad = None, content_type: str = "json") -> Dict[str, Any]:
        """
        makes a request - creates and refresh a token and sessions as needed

        parameters:
            http_method: 'get', 'patch', 'post', 'put', 'delete'
            request_url: API Url, without the common api prefix
            payload : a dictionary , a criteria object, or bytes (for file uploads)
            content_type: any valid content type like json, octet-stream, ...


        :returns
            response_dict: dictionary with the response as dict


        >>> # Setup
        >>> my_api_client = Shopware6AdminAPIClientBase()

        >>> # test resource owner token
        >>> ignore = my_api_client._get_access_token_by_user_credentials()
        >>> my_api_client._get_session()
        >>> ignore = my_api_client._make_request('get', 'customer-group')  # noqa

        >>> # test resource owner token refresh
        >>> my_access_token = my_api_client.token['access_token']
        >>> my_api_client.token['expires_in']=-1
        >>> my_api_client.token['expires_at']=time.time()-1
        >>> ignore = my_api_client._make_request('get', 'customer-group')
        >>> assert my_api_client.token['access_token'] != my_access_token

        >>> # Test client credentials token
        >>> ignore = my_api_client._get_access_token_by_resource_owner()
        >>> my_api_client._get_session()
        >>> ignore = my_api_client._make_request('get', 'customer-group')  # noqa

        >>> # test client credentials token refresh
        >>> my_access_token = my_api_client.token['access_token']
        >>> my_api_client.token['expires_in']=-1
        >>> my_api_client.token['expires_at']=time.time()-1
        >>> ignore = my_api_client._make_request('get', 'customer-group')
        >>> assert my_api_client.token['access_token'] != my_access_token

        >>> # test invalid endpoint
        >>> ignore = my_api_client._make_request('get', 'not-existing')
        Traceback (most recent call last):
            ...
        conf_shopware6_api_base_classes.ShopwareAPIError: 404 Client Error: ...

        """

        retry = 2
        while True:
            try:
                self._get_session()
                response = self._request(http_method=http_method, request_url=request_url, payload=payload, content_type=content_type)
                retry = 0
            except requests_oauthlib.TokenUpdated as exc:
                self._token_saver(token=exc.token)
                response = self._request(http_method=http_method, request_url=request_url, payload=payload, content_type=content_type)
                retry = 0
            except TokenExpiredError:
                if self._is_refreshable_token():  # pragma: no cover
                    # this actually should never happen - just in case.
                    logger.warning("something went wrong - the token should have been automatically refreshed. getting a new token")  # pragma: no cover
                    self._get_access_token_by_user_credentials()  # pragma: no cover
                else:
                    self._get_access_token_by_resource_owner()
                self._get_session()
                response = self._request(http_method=http_method, request_url=request_url, payload=payload, content_type=content_type)
                retry = 0
            except ShopwareAPIError as exc:
                """
                retry   : how often to retry - sometimes we get error code:9, status:401, The resource owner or authorization server denied the request,
                detail: Access token could not be verified.
                But it works if You try again, it seems to be an error in shopware API or race condition
                """
                retry = retry - 1
                if not retry:
                    raise exc

            if not retry:
                break

        try:
            # noinspection PyUnboundLocalVariable
            response_dict = dict(response.json())
        except Exception:  # noqa
            response_dict = dict()
        return response_dict

    def _request(self, http_method: str, request_url: str, payload: PayLoad, content_type: str = "json") -> requests.Response:
        """
        makes a request, needs a self.session to be set up and authenticated

        parameters:
            http_method: 'get', 'patch', 'post', 'put', 'delete'
            request_url: API Url, without the common api prefix
            payload : a dictionary , a criteria object, or bytes (for file uploads)
            content_type: any valid content type like json, octet-stream, ...

        :returns
            response_dict: dictionary with the response as dict

        see : https://docs.python-requests.org/en/latest/user/quickstart/

        :param http_method:
        :param request_url:
        :param payload:
        :return:
        """
        request_data: Union[str, PayLoad]
        payload_dict = dict()

        if _is_type_bytes(payload):
            request_data = payload
            if content_type.lower() == "json":
                raise ShopwareAPIError('Content type "json" does not match the payload data type "bytes"')
        else:
            payload_dict = _get_payload_dict(payload)
            request_data = json.dumps(payload_dict)

        response: requests.Response = requests.Response()
        headers = self._get_headers(content_type=content_type)

        if http_method == "get":
            response = self.session.get(self._format_admin_api_url(request_url), params=payload_dict, headers=headers)
        elif http_method == "patch":
            response = self.session.patch(self._format_admin_api_url(request_url), data=request_data, headers=headers)
        elif http_method == "post":
            response = self.session.post(self._format_admin_api_url(request_url), data=request_data, headers=headers)
        elif http_method == "put":
            response = self.session.put(self._format_admin_api_url(request_url), data=request_data, headers=headers)
        elif http_method == "delete":
            response = self.session.delete(self._format_admin_api_url(request_url))

        try:
            response.raise_for_status()
        except Exception as exc:
            if hasattr(exc, "response"):
                detailed_error = f" : {exc.response.text}"  # type: ignore
            else:
                detailed_error = ""
            raise ShopwareAPIError(f"{exc}{detailed_error}")

        return response

    def _get_token(self) -> Dict[str, Any]:
        """
        get access token, "Client Credentials Grant Type" or "Resource Owner Password Grant Type"

        :parameters
            conf.is_grant_type_resource_owner

        :return:
            the token, also saved in self.token

        >>> # Setup
        >>> my_api_client = Shopware6AdminAPIClientBase()
        >>> save_conf = my_api_client.config.grant_type

        >>> # test "Client Credentials Grant Type"
        >>> my_api_client.config.grant_type = 'user_credentials'
        >>> my_api_client._get_token()
        {'token_type': 'Bearer', 'expires_in': ..., 'access_token': '...', 'refresh_token': '...', 'expires_at': ...}

        >>> # test "Resource Owner Password Grant Type"
        >>> my_api_client.config.grant_type = 'resource_owner'
        >>> my_api_client._get_token()
        {'token_type': 'Bearer', 'expires_in': ..., 'access_token': '...', 'expires_at': ...}

        >>> # Teardown
        >>> my_api_client.config.grant_type = save_conf

        """

        if self.config.grant_type == "user_credentials":
            token = self._get_access_token_by_user_credentials()
        elif self.config.grant_type == "resource_owner":
            token = self._get_access_token_by_resource_owner()
        else:
            raise ShopwareAPIError(f'config.grant_type must bei either "user_credentials" or "resource_owner" not "{str(self.config.grant_type)}"')
        return token

    def _get_access_token_by_resource_owner(self) -> Dict[str, Any]:
        """
        get access token, "Client Credentials Grant Type"
        - no refresh token
        - should be used for machine-to-machine communications, such as CLI jobs or automated services

        see https://shopware.stoplight.io/docs/admin-api/ZG9jOjEwODA3NjQx-authentication-and-authorisation
        setup via Web Administration Interface > settings > system > integration: "access_id" and "access_secret"
        or directly via URL : https://shop.yourdomain.com/admin#/sw/integration/index

        :parameter
            self.shopware_api_url   the api url, like : 'https://shop.yourdomain.com/api'
            self.client_id          the client ID, setup via Web Administration Interface > settings > system > integration: "access_id"
                                    or directly via URL : https://shop.yourdomain.com/admin#/sw/integration/index

            self.client_secret      the client secret, setup via Web Administration Interface > settings > system > integration: "access_secret"
                                    or directly via URL : https://shop.yourdomain.com/admin#/sw/integration/index

        :returns
            self.token

        >>> # Setup
        >>> my_api_client = Shopware6AdminAPIClientBase()
        >>> save_shopware_admin_api_url = my_api_client.config.shopware_admin_api_url
        >>> save_client_id = my_api_client.config.client_id
        >>> save_client_secret = my_api_client.config.client_secret

        >>> # Test Ok
        >>> my_api_client._get_access_token_by_resource_owner()
        {'token_type': 'Bearer', 'expires_in': ..., 'access_token': '...', 'expires_at': ...}

        >>> # Test no url
        >>> my_api_client.config.shopware_admin_api_url = ''
        >>> my_api_client._get_access_token_by_resource_owner()
        Traceback (most recent call last):
            ...
        conf_shopware6_api_base_classes.ShopwareAPIError: shopware_api_url needed
        >>> my_api_client.config.shopware_admin_api_url = save_shopware_admin_api_url

        >>> # Test no client_id
        >>> my_api_client.config.client_id = ''
        >>> my_api_client._get_access_token_by_resource_owner()
        Traceback (most recent call last):
            ...
        conf_shopware6_api_base_classes.ShopwareAPIError: client_id needed
        >>> my_api_client.config.client_id = save_client_id

        >>> # Test no client_secret
        >>> my_api_client.config.client_secret = ''
        >>> my_api_client._get_access_token_by_resource_owner()
        Traceback (most recent call last):
            ...
        conf_shopware6_api_base_classes.ShopwareAPIError: client_secret needed
        >>> my_api_client.config.client_secret = save_client_secret

        """
        if not self.config.shopware_admin_api_url:
            raise ShopwareAPIError("shopware_api_url needed")
        if not self.config.client_id:
            raise ShopwareAPIError("client_id needed")
        if not self.config.client_secret:
            raise ShopwareAPIError("client_secret needed")

        additional_parameters = {"grant_type": "user_credentials"}
        client = oauthlib.oauth2.BackendApplicationClient(client_id=self.config.client_id)
        oauth = requests_oauthlib.OAuth2Session(client=client)
        self.token = oauth.fetch_token(
            token_url=self._format_admin_api_url("oauth/token"),
            client_id=self.config.client_id,
            client_secret=self.config.client_secret,
            kwargs=additional_parameters,
        )
        return self.token

    def _get_access_token_by_user_credentials(self) -> Dict[str, Any]:
        """
        get access token, Integration (Resource Owner Password Grant)
        - with refresh token
        - we recommend to only use this grant flow for client applications that should
          perform administrative actions and require a user-based authentication

        see : https://requests-oauthlib.readthedocs.io/en/latest/oauth2_workflow.html#legacy-application-flow
        see : https://shopware.stoplight.io/docs/admin-api/ZG9jOjEwODA3NjQx-authentication-and-authorisation
        setup at admin/settings/system/user: "access_id" and "access_secret"

        :parameter
            self.shopware_api_url   the api url, like : 'https://shop.yourdomain.com/api'
            self.username           the username, set up at setup at admin/settings/system/users
            self.password           the password, set up at setup at admin/settings/system/users

        :returns
            self.token


        >>> # Setup
        >>> my_api_client = Shopware6AdminAPIClientBase()
        >>> save_shopware_admin_api_url = my_api_client.config.shopware_admin_api_url
        >>> save_username = my_api_client.config.username
        >>> save_password = my_api_client.config.password

        >>> # Test Ok
        >>> my_api_client._get_access_token_by_user_credentials()
        {'token_type': 'Bearer', 'expires_in': ..., 'access_token': '...', 'refresh_token': '...', 'expires_at': ...}

        >>> # Test no url
        >>> my_api_client.config.shopware_admin_api_url = ''
        >>> my_api_client._get_access_token_by_user_credentials()
        Traceback (most recent call last):
            ...
        conf_shopware6_api_base_classes.ShopwareAPIError: shopware_api_url needed
        >>> my_api_client.config.shopware_admin_api_url = save_shopware_admin_api_url

        >>> # Test no username
        >>> my_api_client.config.username = ''
        >>> my_api_client._get_access_token_by_user_credentials()
        Traceback (most recent call last):
            ...
        conf_shopware6_api_base_classes.ShopwareAPIError: username needed
        >>> my_api_client.config.username = save_username

        >>> # Test no password
        >>> my_api_client.config.password = ''
        >>> my_api_client._get_access_token_by_user_credentials()
        Traceback (most recent call last):
            ...
        conf_shopware6_api_base_classes.ShopwareAPIError: password needed
        >>> my_api_client.config.password = save_password

        """

        if not self.config.shopware_admin_api_url:
            raise ShopwareAPIError("shopware_api_url needed")
        if not self.config.username:
            raise ShopwareAPIError("username needed")
        if not self.config.password:
            raise ShopwareAPIError("password needed")

        client_id = "administration"
        additional_parameters = {"grant_type": "password", "scopes": "write"}
        client = oauthlib.oauth2.LegacyApplicationClient(client_id=client_id)
        session_oauth = requests_oauthlib.OAuth2Session(client=client)
        self.token = session_oauth.fetch_token(
            token_url=self._format_admin_api_url("oauth/token"),
            client_id=client_id,
            username=self.config.username,
            password=self.config.password,
            kwargs=additional_parameters,
        )
        return self.token

    def _get_session(self) -> None:
        """
        see : https://requests-oauthlib.readthedocs.io/en/latest/oauth2_workflow.html#legacy-application-flow
        see : https://shopware.stoplight.io/docs/admin-api/ZG9jOjEwODA3NjQx-authentication-and-authorisation


        >>> # Setup
        >>> my_api_client = Shopware6AdminAPIClientBase()

        >>> # Test client credentials token
        >>> ignore = my_api_client._get_access_token_by_resource_owner()
        >>> my_api_client._get_session()
        >>> data = my_api_client.request_get('customer-group')

        >>> # Test resource owner token
        >>> ignore = my_api_client._get_access_token_by_user_credentials()
        >>> my_api_client._get_session()
        >>> data = my_api_client.request_get('customer-group')

        >>> # Test resource owner Token Refresh
        >>> my_access_token = my_api_client.token['access_token']
        >>> my_api_client.token['expires_in']=-1
        >>> my_api_client.token['expires_at']=time.time()-1
        >>> my_api_client._get_session()
        >>> data = my_api_client.request_get('customer-group')
        >>> assert my_api_client.token['access_token'] != my_access_token

        """
        try:
            if not self.token:
                self._get_token()
            if self._is_refreshable_token():
                self.token["expires_in"] = int(self.token["expires_at"] - time.time())
                client_id = "administration"
                extra = {"client_id": client_id}
                self.session = requests_oauthlib.OAuth2Session(
                    client_id, token=self.token, auto_refresh_kwargs=extra, auto_refresh_url=self._format_admin_api_url("oauth/token")
                )
            else:
                client_id = self.config.client_id
                self.session = requests_oauthlib.OAuth2Session(client_id, token=self.token)
        except requests_oauthlib.TokenUpdated as exc:
            self._token_saver(token=exc.token)

    def _token_saver(self, token: Dict[str, str]) -> None:
        """
        saves the token - this is needed for automatically refreshing the "resource owner" access token.
        the "user_credentials" can not be refreshed

        :parameter
            token:             the token to be saved

        :returns
            None

        """
        self.token = token

    def _is_refreshable_token(self) -> bool:
        """
        True if the token is refreshable ().
        the "resource owner" access token can be refreshed
        the "client_credentials" can not be refreshed
        """
        return "refresh_token" in self.token

    def _format_admin_api_url(self, request_url: str) -> str:
        """
        formatted url to make a request

        :parameter
            request_url:             the request url, for instance "oauth/token"
            self.shopware_api_url:          the api url, for instance https://your.shop-domain.com/api

        :returns
            the formatted url, like  https://your.shop-domain.com/api/oauth/token

        >>> my_api_client = Shopware6AdminAPIClientBase()
        >>> my_api_client._format_admin_api_url('test')
        'http.../api/test'

        """
        request_url = request_url.lstrip("/")
        return f"{self.config.shopware_admin_api_url}/{request_url}"

    @lru_cache(maxsize=None)
    def _get_headers(self, content_type: str = "json") -> Dict[str, str]:
        """
        content_type can be any valid content type like json, octet-stream,

        >>> my_api_client = Shopware6AdminAPIClientBase()
        >>> my_api_client._get_headers()    # noqa
        {'Content-Type': 'application/json', 'Accept': 'application/json'}

        >>> my_api_client._get_headers(content_type='json')    # noqa
        {'Content-Type': 'application/json', 'Accept': 'application/json'}

        >>> my_api_client._get_headers(content_type='octet-stream')    # noqa
        {'Content-Type': 'application/octet-stream', 'Accept': 'application/json'}

        >>> # Teardown
        >>> my_api_client._get_headers.cache_clear()    # noqa


        """
        headers = {"Content-Type": f"application/{content_type.lower()}", "Accept": "application/json"}
        return headers


def _load_config(use_docker_test_container: bool) -> ConfShopware6ApiBase:
    if _is_github_actions() or use_docker_test_container:  # pragma: no cover
        config = _load_config_for_docker_test_container()
        config.store_api_sw_access_key = _get_docker_test_container_store_access_key()
        _create_docker_test_container_resource_owner_credentials()
    else:
        config = _load_config_for_rotek_production()  # pragma: no cover
    return config


def _load_config_for_docker_test_container() -> ConfShopware6ApiBase:
    try:
        from conf_shopware6_api_base_docker_testcontainer import conf_shopware6_api_base
    except ImportError:  # pragma: no cover
        # Imports for Doctest
        from .conf_shopware6_api_base_docker_testcontainer import conf_shopware6_api_base  # type: ignore  # pragma: no cover
    return conf_shopware6_api_base  # type: ignore


def _get_docker_test_container_store_access_key() -> str:
    """
    get the sales channel accessKey from the docker testcontainer
    :return:

    >>> if _is_local_docker_container_active():
    ...     store_access_key = _get_docker_test_container_store_access_key()
    ...     assert 26 == len(store_access_key)

    """
    config = _load_config_for_docker_test_container()
    admin_api_client = Shopware6AdminAPIClientBase(config=config)
    admin_api_client._get_access_token_by_user_credentials()
    admin_api_client._get_session()
    response_dict = admin_api_client.request_get("sales-channel")
    access_key = str(response_dict["data"][0]["accessKey"])
    return access_key


def _create_docker_test_container_resource_owner_credentials() -> None:
    """
    >>> if _is_local_docker_container_active():
    ...     _create_docker_test_container_resource_owner_credentials()
    ...     assert True == _is_resource_owner_credentials_present()

    """
    if not _is_resource_owner_credentials_present():
        _upsert_docker_test_container_resource_owner_credentials()


def _upsert_docker_test_container_resource_owner_credentials() -> None:
    """
    inserts resource_owner_credentials (admin) on the docker test container
    the credentials are hardcoded here and are also reflected in the config file

    """
    payload = {
        "id": "565c4ada878141d3b18d6977dbbd2a13",  # noqa
        "label": "dockware_integration_admin",  # noqa
        "accessKey": "SWIACWJOMUTXV1RMNGJUAKTUAA",  # noqa
        "secretAccessKey": "UkhvUG1qdmpuMjFudGJCdG1Xc0xMbEt2ck9CQ2xDTUtXMUZHRUQ",  # noqa
        "admin": True,
    }  # noqa
    config = _load_config_for_docker_test_container()
    admin_api_client = Shopware6AdminAPIClientBase(config=config)
    admin_api_client._get_access_token_by_user_credentials()
    admin_api_client._get_session()
    admin_api_client.request_post("integration", payload=payload)


def _is_resource_owner_credentials_present() -> bool:
    """
    Returns True if there are resource owner credentials present

    >>> if _is_local_docker_container_active():
    ...     discard = _is_resource_owner_credentials_present()

    """
    config = _load_config_for_docker_test_container()
    admin_api_client = Shopware6AdminAPIClientBase(config=config)
    admin_api_client._get_access_token_by_user_credentials()
    admin_api_client._get_session()
    response_dict = admin_api_client.request_post("search/integration")
    if response_dict["total"]:
        resource_owner_credentials_present = True
    else:
        resource_owner_credentials_present = False
    return resource_owner_credentials_present


def _load_config_for_rotek_production() -> ConfShopware6ApiBase:
    """
    load config file for the rotek shop

    """
    try:  # pragma: no cover
        from conf_shopware6_api_base_rotek import conf_shopware6_api_base  # pragma: no cover
    except ImportError:  # pragma: no cover
        # Imports for Doctest
        from .conf_shopware6_api_base_rotek import conf_shopware6_api_base  # type: ignore  # pragma: no cover
    return conf_shopware6_api_base  # type: ignore


def _is_github_actions() -> bool:
    """
    True if run on github actions
    >>> discard = _is_github_actions()
    """
    return os.getenv("GITHUB_ACTIONS", "false").lower() == "true"


def _is_local_docker_container_active() -> bool:
    """
    True if the local docker container is running

    >>> discard = _is_local_docker_container_active()

    """
    try:
        requests.get("http://localhost/admin")
        is_active = True
    except requests.exceptions.ConnectionError:  # pragma: no cover
        is_active = False  # pragma: no cover
    return is_active


def _is_type_bytes(payload: PayLoad) -> bool:
    """True if the passed type is bytes"""
    return type(payload).__name__ == "bytes"


def _is_type_criteria(payload: PayLoad) -> bool:
    """True if the passed type is Criteria"""
    return type(payload).__name__ == "Criteria"


def _get_payload_dict(payload: PayLoad) -> Dict[str, Any]:
    """return the payload as dictionary"""
    if payload is None:
        payload = dict()
    elif _is_type_criteria(payload):
        payload = payload.get_dict()  # type: ignore
    return payload  # type: ignore


if __name__ == "__main__":
    print(b'this is a library only, the executable is named "lib_shopware6_api_base_cli.py"', file=sys.stderr)



# STDLIB
import sys
from typing import Optional

# EXT
import click

# OWN
import cli_exit_tools

# PROJ
try:
    from . import __init__conf__
    from . import lib_shopware6_api_base
except (ImportError, ModuleNotFoundError):  # pragma: no cover
    # imports for doctest
    import __init__conf__  # type: ignore  # pragma: no cover
    import lib_shopware6_api_base  # type: ignore  # pragma: no cover

# CONSTANTS
CLICK_CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


def info() -> None:
    """
    >>> info()
    Info for ...

    """
    __init__conf__.print_info()


@click.group(help=__init__conf__.title, context_settings=CLICK_CONTEXT_SETTINGS)
@click.version_option(
    version=__init__conf__.version, prog_name=__init__conf__.shell_command, message=f"{__init__conf__.shell_command} version {__init__conf__.version}"
)
@click.option("--traceback/--no-traceback", is_flag=True, type=bool, default=None, help="return traceback information on cli")
def cli_main(traceback: Optional[bool] = None) -> None:
    if traceback is not None:
        cli_exit_tools.config.traceback = traceback


@cli_main.command("info", context_settings=CLICK_CONTEXT_SETTINGS)  # type: ignore
def cli_info() -> None:
    """get program informations"""
    info()


# entry point if main
if __name__ == "__main__":
    try:
        cli_main()
    except Exception as exc:
        cli_exit_tools.print_exception_message()
        sys.exit(cli_exit_tools.get_system_exit_code(exc))
    finally:
        cli_exit_tools.flush_streams()



# STDLIB
from typing import Optional, Union

# EXT
import attrs

# proj
try:
    from lib_shopware6_api_base_helpers import get_pretty_printer
except ImportError:  # pragma: no cover
    # Imports for Doctest
    from .lib_shopware6_api_base_helpers import get_pretty_printer


# FieldSorting{{{
@attrs.define
class FieldSorting:
    """
    see: https://shopware.stoplight.io/docs/store-api/ZG9jOjEwODExNzU2-search-queries#sort
    The sort parameter allows to control the sorting of the result. Several sorts can be transferred at the same time.
    The field parameter defines which field is to be used for sorting.
    The order parameter defines the sort direction.
    The parameter naturalSorting allows to use a Natural Sorting Algorithm

    :parameter
        field : str
        order : str "ASC" or "DESC"
        naturalSorting : Optional[bool]

    >>> # Setup
    >>> pp = get_pretty_printer()

    >>> # Test
    >>> my_sorting = FieldSorting('name', 'ASC', True)
    >>> pp(attrs.asdict(my_sorting))
    {'field': 'name', 'order': 'ASC', 'naturalSorting': True}

    """

    # FieldSorting}}}
    field: str = attrs.field(validator=attrs.validators.instance_of(str))
    order: str = attrs.field(validator=attrs.validators.in_(["ASC", "DESC"]))
    naturalSorting: Optional[bool] = None


# AscFieldSorting{{{
@attrs.define
class AscFieldSorting:
    """
    see: https://shopware.stoplight.io/docs/store-api/ZG9jOjEwODExNzU2-search-queries#sort
    The sort parameter allows to control the sorting of the result. Several sorts can be transferred at the same time.
    The field parameter defines which field is to be used for sorting.
    The order parameter defines the sort direction.
    The parameter naturalSorting allows to use a Natural Sorting Algorithm

    :parameter
        field : str
        naturalSorting : Optional[bool]

    >>> # Setup
    >>> pp = get_pretty_printer()

    >>> # Test
    >>> my_sorting = AscFieldSorting('name', True)
    >>> pp(attrs.asdict(my_sorting))
    {'field': 'name', 'order': 'ASC', 'naturalSorting': True}

    """

    # AscFieldSorting}}}

    field: str = attrs.field(validator=attrs.validators.instance_of(str))
    order: str = attrs.field(init=False, default="ASC")
    naturalSorting: Optional[bool] = None


# DescFieldSorting{{{
@attrs.define
class DescFieldSorting:
    """
    see: https://shopware.stoplight.io/docs/store-api/ZG9jOjEwODExNzU2-search-queries#sort
    The sort parameter allows to control the sorting of the result. Several sorts can be transferred at the same time.
    The field parameter defines which field is to be used for sorting.
    The order parameter defines the sort direction.
    The parameter naturalSorting allows to use a Natural Sorting Algorithm

    :parameter
        field : str
        naturalSorting : Optional[bool]

    >>> # Setup
    >>> pp = get_pretty_printer()

    >>> # Test
    >>> my_sorting = DescFieldSorting('name', True)
    >>> pp(attrs.asdict(my_sorting))
    {'field': 'name', 'order': 'DESC', 'naturalSorting': True}

    """

    # DescFieldSorting}}}

    field: str = attrs.field(validator=attrs.validators.instance_of(str))
    order: str = attrs.field(init=False, default="DESC")
    naturalSorting: Optional[bool] = None


SortType = Union[FieldSorting, AscFieldSorting, DescFieldSorting]



# STDLIB
from datetime import datetime
import sys
from typing import Any, Dict, List, Optional, Union

# OWN
import lib_detect_testenv

# EXT
import attrs
from attrs import validators


# proj
try:
    from lib_shopware6_api_base_helpers import get_pretty_printer
    from lib_shopware6_api_base_criteria_aggregation import *
    from lib_shopware6_api_base_criteria_filter import *
    from lib_shopware6_api_base_criteria_sorting import *
except ImportError:  # pragma: no cover
    # Imports for Doctest
    from .lib_shopware6_api_base_helpers import get_pretty_printer
    from .lib_shopware6_api_base_criteria_aggregation import *  # type: ignore  # pragma: no cover
    from .lib_shopware6_api_base_criteria_filter import *  # type: ignore  # pragma: no cover
    from .lib_shopware6_api_base_criteria_sorting import *  # type: ignore  # pragma: no cover


PostFilterType = "T"  # not implemented now


# Query{{{
@attrs.define
class Query:
    """
    see: https://shopware.stoplight.io/docs/store-api/ZG9jOjEwODExNzU2-search-queries#query
    Enables you to determine a ranking for the search result
    Use this parameter to create a weighted search query that returns a _score for each found entity.
    Any filter type can be used for the query. A score has to be defined for each query.
    The sum of the matching queries then results in the total _score value.

    :parameter
        score   int
        query   FilterType

    >>> # Setup
    >>> pp = get_pretty_printer()

    >>> # Test
    >>> my_criteria = Criteria(
    ...    query=[Query(score=500, query=ContainsFilter(field='name', value='Bronze')),
    ...           Query(score=500, query=EqualsFilter(field='active', value='true')),
    ...           Query(score=100, query=EqualsFilter(field='manufacturerId', value='db3c17b1e572432eb4a4c881b6f9d68f'))])
    >>> pp(my_criteria.get_dict())
    {'limit': None,
     'page': None,
     'query': [{'score': 500,
                'query': {'type': 'contains', 'field': 'name', 'value': 'Bronze'}},
               {'score': 500,
                'query': {'type': 'equals', 'field': 'active', 'value': 'true'}},
               {'score': 100,
                'query': {'type': 'equals',
                          'field': 'manufacturerId',
                          'value': 'db3c17b1e572432eb4a4c881b6f9d68f'}}],
     'term': None,
     'total_count_mode': None}

    """

    # Query}}}
    score: int
    query: FilterType


# criteria{{{
@attrs.define
class Criteria:
    """
    see: https://shopware.stoplight.io/docs/store-api/ZG9jOjEwODExNzU2-search-queries

    structure of Criteria:

    parameter:

    aggregations  List[Aggregation]                    Specify aggregations to be computed on-the-fly
    associations  Dict['<name>', 'Criteria']           Allows to load additional data to the standard data of an entity
    filter        List[Filter]                         Allows you to filter the result and aggregations
    grouping      List['<fieldname>']                  allows you to group the result over fields
    ids           List['<id>']                         Limits the search to a list of Ids
    includes      Dict['apiAlias', List[<fieldname>]]  Restricts the output to the defined fields
    limit         Optional[int]                        Defines the number of entries to be determined
    page          Optional[int]                        Defines at which page the search result should start
    post-filter                           not implemented at the moment
    query         List[Query]                          Enables you to determine a ranking for the search result
    sort          List[Sort]                           Defines the sorting of the search result
    term          Optional[str]                        text search on all records based on their data model and weighting
                                                       Don't use term parameters together with query parameters.
    total-count-mode    Optional[int]                  Defines whether a total must be determined



    >>> # Setup
    >>> pp = get_pretty_printer()

    >>> # Test empty
    >>> my_criteria = Criteria()
    >>> pp(my_criteria.get_dict())
    {'limit': None, 'page': None, 'term': None, 'total_count_mode': None}

    >>> # Test Average aggregation
    >>> my_criteria = Criteria()
    >>> my_criteria.limit=1
    >>> my_criteria.includes['product'] = ['id', 'name']
    >>> my_criteria.aggregations = [AvgAggregation('average-price', 'price')]
    >>> pp(my_criteria.get_dict())
    {'limit': 1,
     'page': None,
     'aggregations': [{'name': 'average-price', 'type': 'avg', 'field': 'price'}],
     'includes': {'product': ['id', 'name']},
     'term': None,
     'total_count_mode': None}

    >>> # Test Filter aggregation
    >>> my_criteria = Criteria(limit=1, includes={'product':['id', 'name']},
    ...     aggregations=FilterAggregation(name='active-price-avg',
    ...                                    filter=EqualsFilter(field='active', value=True),
    ...                                    aggregation=AvgAggregation(name='avg-price',field='price')))
    >>> pp(my_criteria.get_dict())
    {'limit': 1,
     'page': None,
     'aggregations': {'name': 'active-price-avg',
                      'type': 'filter',
                      'filter': {'type': 'equals',
                                 'field': 'active',
                                 'value': True},
                      'aggregation': {'name': 'avg-price',
                                      'type': 'avg',
                                      'field': 'price'}},
     'includes': {'product': ['id', 'name']},
     'term': None,
     'total_count_mode': None}

    >>> # Association{{{
    >>> # Test Association
    >>> my_criteria = Criteria()
    >>> my_criteria.associations['products'] = Criteria(limit=5, filter=[EqualsFilter('active', 'true')])
    >>> pp(my_criteria.get_dict())
    {'limit': None,
     'page': None,
     'associations': {'products': {'limit': 5,
                                   'page': None,
                                   'filter': [{'type': 'equals',
                                               'field': 'active',
                                               'value': 'true'}],
                                   'term': None,
                                   'total_count_mode': None}},
     'term': None,
     'total_count_mode': None}
    >>> # Association}}}

    >>> # Test append filters
    >>> my_criteria = Criteria()
    >>> my_criteria.page = 0
    >>> my_criteria.limit=1
    >>> my_criteria.filter.append(EqualsFilter('a', 'a'))
    >>> my_criteria.filter.append(EqualsFilter('b', 'b'))
    >>> my_criteria.filter.append(EqualsFilter('d', 'd'))
    >>> pp(my_criteria.get_dict())
    {'limit': 1,
     'page': 0,
     'filter': [{'type': 'equals', 'field': 'a', 'value': 'a'},
                {'type': 'equals', 'field': 'b', 'value': 'b'},
                {'type': 'equals', 'field': 'd', 'value': 'd'}],
     'term': None,
     'total_count_mode': None}

    >>> # Test set filters
    >>> my_criteria = Criteria()
    >>> my_criteria.filter = [EqualsFilter('a', 'a'), EqualsFilter('b', 'b'), EqualsFilter('d', 'd')]
    >>> pp(my_criteria.get_dict())
    {'limit': None,
     'page': None,
     'filter': [{'type': 'equals', 'field': 'a', 'value': 'a'},
                {'type': 'equals', 'field': 'b', 'value': 'b'},
                {'type': 'equals', 'field': 'd', 'value': 'd'}],
     'term': None,
     'total_count_mode': None}

    >>> # Grouping{{{
    >>> # Test Grouping
    >>> my_criteria = Criteria()
    >>> my_criteria.limit=5
    >>> my_criteria.grouping=['active']
    >>> pp(my_criteria.get_dict())
    {'limit': 5,
     'page': None,
     'grouping': ['active'],
     'term': None,
     'total_count_mode': None}
    >>> # Grouping}}}

    >>> # ids{{{
    >>> # Test ids
    >>> # note that the limit is automatically set to 3, which is for our paginated request
    >>> my_criteria = Criteria()
    >>> my_criteria.ids=["012cd563cf8e4f0384eed93b5201cc98", "075fb241b769444bb72431f797fd5776", "090fcc2099794771935acf814e3fdb24"]
    >>> pp(my_criteria.get_dict())
    {'limit': 3,
     'page': None,
     'ids': ['012cd563cf8e4f0384eed93b5201cc98',
             '075fb241b769444bb72431f797fd5776',
             '090fcc2099794771935acf814e3fdb24'],
     'term': None,
     'total_count_mode': None}
    >>> # ids}}}

    >>> # includes{{{
    >>> # Test includes
    >>> my_criteria = Criteria()
    >>> my_criteria.includes['product'] = ['id', 'name']
    >>> pp(my_criteria.get_dict())
    {'limit': None,
     'page': None,
     'includes': {'product': ['id', 'name']},
     'term': None,
     'total_count_mode': None}

    >>> # includes}}}

    >>> # page&limit{{{
    >>> my_criteria = Criteria(page=1, limit=5)
    >>> pp(my_criteria.get_dict())
    {'limit': 5, 'page': 1, 'term': None, 'total_count_mode': None}

    >>> # page&limit}}}

    >>> # Test Query
    >>> my_criteria = Criteria(
    ...    query=[Query(score=500, query=ContainsFilter(field='name', value='Bronze')),
    ...           Query(score=500, query=EqualsFilter(field='active', value='true')),
    ...           Query(score=100, query=EqualsFilter(field='manufacturerId', value='db3c17b1e572432eb4a4c881b6f9d68f'))])
    >>> pp(my_criteria.get_dict())
    {'limit': None,
     'page': None,
     'query': [{'score': 500,
                'query': {'type': 'contains', 'field': 'name', 'value': 'Bronze'}},
               {'score': 500,
                'query': {'type': 'equals', 'field': 'active', 'value': 'true'}},
               {'score': 100,
                'query': {'type': 'equals',
                          'field': 'manufacturerId',
                          'value': 'db3c17b1e572432eb4a4c881b6f9d68f'}}],
     'term': None,
     'total_count_mode': None}

    >>> # Test Sorting
    >>> my_criteria = Criteria(limit=5,
    ...                        sort=[FieldSorting('name', 'ASC', True),
    ...                              DescFieldSorting('active')])
    >>> pp(my_criteria.get_dict())
    {'limit': 5,
     'page': None,
     'sort': [{'field': 'name', 'order': 'ASC', 'naturalSorting': True},
              {'field': 'active', 'order': 'DESC', 'naturalSorting': None}],
     'term': None,
     'total_count_mode': None}

    """

    # criteria}}}

    limit: Optional[int] = None
    page: Optional[int] = None
    aggregations: List["AggregationType"] = attrs.field(factory=list)
    associations: Dict[str, "Criteria"] = attrs.field(factory=dict)
    filter: List[FilterType] = attrs.field(factory=list)
    grouping: List[str] = attrs.field(factory=list)
    ids: List[str] = attrs.field(factory=list)
    includes: Dict[str, List[str]] = attrs.field(factory=dict)
    post_filter: List["PostFilterType"] = attrs.field(factory=list)  # type: ignore # not implemented now
    query: List[Query] = attrs.field(factory=list)
    sort: List["SortType"] = attrs.field(factory=list)
    term: Optional[str] = None
    total_count_mode: Optional[int] = None

    def get_dict(self) -> Dict[str, Any]:
        result = attrs.asdict(self, filter=_is_not_empty)

        if self.ids:
            if self.limit:
                raise ValueError('You can use either "limit" or "ids", but not both, see : https://github.com/bitranox/lib_shopware6_api_base#ids')
            else:
                result["limit"] = len(self.ids)

        return result


def _is_not_empty(attribute: Any, value: Any) -> bool:
    """Filter out empty Lists and Dictionaries"""
    if value == dict():
        return False
    elif value == list():
        return False
    else:
        return True



# STDLIB
import sys
from typing import Any, Callable


def get_pretty_printer() -> Callable[[Any], str]:
    # this is only needed for Python3.6, Python3.7
    is_before_python38 = (sys.version_info.major, sys.version_info.minor) < (3, 8)
    if is_before_python38:
        import pprint3x as pprint  # type: ignore
    else:
        import pprint  # type: ignore
    pp = pprint.PrettyPrinter(sort_dicts=False).pprint  # type: ignore
    return pp  # type: ignore



# STDLIB
from datetime import datetime
from typing import Any, Dict, List, Optional, Union

# EXT
import attrs
from attrs import validators

# proj
try:
    from lib_shopware6_api_base_helpers import get_pretty_printer
    from lib_shopware6_api_base_criteria_sorting import *
    from lib_shopware6_api_base_criteria_filter import *
except ImportError:  # pragma: no cover
    # Imports for Doctest
    from .lib_shopware6_api_base_helpers import get_pretty_printer
    from .lib_shopware6_api_base_criteria_sorting import *  # type: ignore  # pragma: no cover
    from .lib_shopware6_api_base_criteria_filter import *  # type: ignore  # pragma: no cover


@attrs.define
class _AggregationNames:
    """
    see: https://developer.shopware.com/docs/resources/references/core-reference/dal-reference/aggregations-reference
    definition of Aggregation Names

    """

    avg: str = "avg"
    count: str = "count"
    max: str = "max"
    min: str = "min"
    stats: str = "stats"
    sum: str = "sum"
    filter: str = "filter"
    terms: str = "terms"
    histogram: str = "histogram"


aggregation_names = _AggregationNames()


# AvgAggregation{{{
@attrs.define
class AvgAggregation:
    """
    see aggregations reference : https://developer.shopware.com/docs/resources/references/core-reference/dal-reference/aggregations-reference
    The Avg aggregation makes it possible to calculate the average value for a field.
    The following SQL statement is executed in the background: AVG(price).

    :parameter:
        name: str
        field: str

    >>> # Setup
    >>> pp = get_pretty_printer()

    >>> # Test
    >>> my_aggregation = AvgAggregation('avg-price', 'price')
    >>> pp(attrs.asdict(my_aggregation))
    {'name': 'avg-price', 'type': 'avg', 'field': 'price'}

    """

    # AvgAggregation}}}

    name: str = attrs.field(validator=attrs.validators.instance_of(str))
    type: str = attrs.field(init=False, default="avg")
    field: str = attrs.field(validator=attrs.validators.instance_of(str))


# CountAggregation{{{
@attrs.define
class CountAggregation:
    """
    see aggregations reference : https://developer.shopware.com/docs/resources/references/core-reference/dal-reference/aggregations-reference
    The count aggregation makes it possible to determine the number of entries for a field that are filled with a value.
    The following SQL statement is executed in the background: COUNT(DISTINCT(manufacturerId)).

    :parameter:
        name: str
        field: str

    >>> # Setup
    >>> pp = get_pretty_printer()

    >>> # Test
    >>> my_aggregation = CountAggregation('count-manufacturers', 'manufacturerId')
    >>> pp(attrs.asdict(my_aggregation))
    {'name': 'count-manufacturers', 'type': 'count', 'field': 'manufacturerId'}

    """

    # CountAggregation}}}

    name: str = attrs.field(validator=attrs.validators.instance_of(str))
    type: str = attrs.field(init=False, default="count")
    field: str = attrs.field(validator=attrs.validators.instance_of(str))


# MaxAggregation{{{
@attrs.define
class MaxAggregation:
    """
    see aggregations reference : https://developer.shopware.com/docs/resources/references/core-reference/dal-reference/aggregations-reference
    The max aggregation allows you to determine the maximum value of a field.
    The following SQL statement is executed in the background: MAX(price).

    :parameter:
        name: str
        field: str

    >>> # Setup
    >>> pp = get_pretty_printer()

    >>> # Test
    >>> my_aggregation = MaxAggregation('max-price', 'price')
    >>> pp(attrs.asdict(my_aggregation))
    {'name': 'max-price', 'type': 'max', 'field': 'price'}

    """

    # MaxAggregation}}}

    name: str = attrs.field(validator=attrs.validators.instance_of(str))
    type: str = attrs.field(init=False, default="max")
    field: str = attrs.field(validator=attrs.validators.instance_of(str))


# MinAggregation{{{
@attrs.define
class MinAggregation:
    """
    see aggregations reference : https://developer.shopware.com/docs/resources/references/core-reference/dal-reference/aggregations-reference
    The min aggregation makes it possible to determine the minimum value of a field.
    The following SQL statement is executed in the background: MIN(price)

    :parameter:
        name: str
        field: str

    >>> # Setup
    >>> pp = get_pretty_printer()

    >>> # Test
    >>> my_aggregation = MinAggregation('min-price', 'price')
    >>> pp(attrs.asdict(my_aggregation))
    {'name': 'min-price', 'type': 'min', 'field': 'price'}

    """

    # MinAggregation}}}

    name: str = attrs.field(validator=attrs.validators.instance_of(str))
    type: str = attrs.field(init=False, default="min")
    field: str = attrs.field(validator=attrs.validators.instance_of(str))


# SumAggregation{{{
@attrs.define
class SumAggregation:
    """
    see aggregations reference : https://developer.shopware.com/docs/resources/references/core-reference/dal-reference/aggregations-reference
    The sum aggregation makes it possible to determine the total of a field.
    The following SQL statement is executed in the background: SUM(price)

    :parameter:
        name: str
        field: str

    >>> # Setup
    >>> pp = get_pretty_printer()

    >>> # Test
    >>> my_aggregation = SumAggregation('sum-price', 'price')
    >>> pp(attrs.asdict(my_aggregation))
    {'name': 'sum-price', 'type': 'sum', 'field': 'price'}

    """

    # SumAggregation}}}

    name: str = attrs.field(validator=attrs.validators.instance_of(str))
    type: str = attrs.field(init=False, default="sum")
    field: str = attrs.field(validator=attrs.validators.instance_of(str))


# StatsAggregation{{{
@attrs.define
class StatsAggregation:
    """
    see aggregations reference : https://developer.shopware.com/docs/resources/references/core-reference/dal-reference/aggregations-reference
    The stats aggregation makes it possible to calculate several values at once for a field.
    This includes the previous max, min, avg and sum aggregation.
    The following SQL statement is executed in the background: SELECT MAX(price), MIN(price), AVG(price), SUM(price)

    :parameter:
        name: str
        field: str

    >>> # Setup
    >>> pp = get_pretty_printer()

    >>> # Test
    >>> my_aggregation = StatsAggregation('stats-price', 'price')
    >>> pp(attrs.asdict(my_aggregation))
    {'name': 'stats-price', 'type': 'stats', 'field': 'price'}

    """

    # StatsAggregation}}}

    name: str = attrs.field(validator=attrs.validators.instance_of(str))
    type: str = attrs.field(init=False, default="stats")
    field: str = attrs.field(validator=attrs.validators.instance_of(str))


# TermsAggregation{{{
@attrs.define
class TermsAggregation:
    """
    see aggregations reference : https://developer.shopware.com/docs/resources/references/core-reference/dal-reference/aggregations-reference

    The terms aggregation belongs to the bucket aggregations.
    This allows you to determine the values of a field.
    The result contains each value once and how often this value occurs in the result.
    The terms aggregation also supports the following parameters:
        limit - Defines a maximum number of entries to be returned (default: zero)
        sort - Defines the order of the entries. By default the following is not sorted
        aggregation - Enables you to calculate further aggregations for each key
    The following SQL statement is executed in the background: SELECT DISTINCT(manufacturerId) as key, COUNT(manufacturerId) as count


    :parameter:
        name: str
        field: str
        sort: Optional[SortType]
        limit: Optional[int]
        aggregation: Optional[]

    >>> # Setup
    >>> pp = get_pretty_printer()

    >>> # Test
    >>> my_aggregation = TermsAggregation(name='manufacturer-ids', limit=3, sort=DescFieldSorting('manufacturer.name'), field='manufacturerId')
    >>> pp(attrs.asdict(my_aggregation))
    {'name': 'manufacturer-ids',
     'type': 'terms',
     'field': 'manufacturerId',
     'sort': {'field': 'manufacturer.name',
              'order': 'DESC',
              'naturalSorting': None},
     'limit': 3,
     'aggregation': None}

    """

    # TermsAggregation}}}

    name: str = attrs.field(validator=attrs.validators.instance_of(str))
    type: str = attrs.field(init=False, default="terms")
    field: str = attrs.field(validator=attrs.validators.instance_of(str))
    sort: Optional[SortType] = None
    limit: Optional[int] = attrs.field(validator=attrs.validators.instance_of(int), default=None)
    aggregation: Optional["AggregationType"] = None


# FilterAggregation{{{
@attrs.define
class FilterAggregation:
    """
    see aggregations reference : https://developer.shopware.com/docs/resources/references/core-reference/dal-reference/aggregations-reference

    The filter aggregation belongs to the bucket aggregations.
    Unlike all other aggregations, this aggregation does not determine any result, it cannot be used alone.
    It is only used to further restrict the result of an aggregation in a criterion.
    Filters which defined inside the filter property of this aggregation type,
    are only used when calculating this aggregation.
    The filters have no effect on other aggregations or on the result of the search.

    :parameter:
        name: str
        sort: SortType
        filter: FilterType
        aggregation : AggregationType

    >>> # Setup
    >>> pp = get_pretty_printer()

    >>> # Test
    >>> my_aggregation = FilterAggregation(
    ...     name='active-price-avg',
    ...     filter=EqualsFilter(field='active', value=True),
    ...     aggregation=AvgAggregation(name='avg-price',field='price'))
    >>> pp(attrs.asdict(my_aggregation))
    {'name': 'active-price-avg',
     'type': 'filter',
     'filter': {'type': 'equals', 'field': 'active', 'value': True},
     'aggregation': {'name': 'avg-price', 'type': 'avg', 'field': 'price'}}

    """

    # FilterAggregation}}}

    name: str = attrs.field(validator=attrs.validators.instance_of(str))
    type: str = attrs.field(init=False, default="filter")
    filter: FilterType
    aggregation: "AggregationType"


# EntityAggregation{{{
@attrs.define
class EntityAggregation:
    """
    see aggregations reference : https://developer.shopware.com/docs/resources/references/core-reference/dal-reference/aggregations-reference

    The entity aggregation is similar to the terms aggregation, it belongs to the bucket aggregations.
    As with terms aggregation, all unique values are determined for a field.
    The aggregation then uses the determined keys to load the defined entity. The keys are used here as ids.

    :parameter:
        name: str
        definition: str
        field: str

    >>> # Setup
    >>> pp = get_pretty_printer()

    >>> # Test
    >>> my_aggregation = EntityAggregation(name='manufacturers', definition='product_manufacturer', field='manufacturerId')
    >>> pp(attrs.asdict(my_aggregation))
    {'name': 'manufacturers',
     'type': 'entity',
     'definition': 'product_manufacturer',
     'field': 'manufacturerId'}
    """

    # EntityAggregation}}}

    name: str = attrs.field(validator=attrs.validators.instance_of(str))
    type: str = attrs.field(init=False, default="entity")
    definition: str = attrs.field(validator=attrs.validators.instance_of(str))
    field: str = attrs.field(validator=attrs.validators.instance_of(str))


# DateHistogramAggregation{{{
@attrs.define
class DateHistogramAggregation:
    """
    see aggregations reference : https://developer.shopware.com/docs/resources/references/core-reference/dal-reference/aggregations-reference

    The histogram aggregation is used as soon as the data to be determined refers to a date field.
    With the histogram aggregation,
    one of the following date intervals can be given: minute, hour, day, week, month, quarter, year, day.
    This interval groups the result and calculates the corresponding count of hits.

    :parameter:
        name: str
        field: str
        interval: str ,  possible values: 'minute', 'hour', 'day', 'week', 'month', 'quarter', 'year', 'day'

    >>> # Setup
    >>> pp = get_pretty_printer()

    >>> # Test
    >>> my_aggregation = DateHistogramAggregation(name='release-dates', field='releaseDate', interval='month')
    >>> pp(attrs.asdict(my_aggregation))
    {'name': 'release-dates',
     'type': 'histogram',
     'field': 'releaseDate',
     'interval': 'month'}

    """

    # DateHistogramAggregation}}}

    name: str = attrs.field(validator=attrs.validators.instance_of(str))
    type: str = attrs.field(init=False, default="histogram")
    field: str = attrs.field(validator=attrs.validators.instance_of(str))
    interval: str = attrs.field(validator=attrs.validators.in_(["minute", "hour", "day", "week", "month", "quarter", "year", "day"]))


# NestingAggregations{{{
"""
see: https://developer.shopware.com/docs/resources/references/core-reference/dal-reference/aggregations-reference#nesting-aggregations
"""
# NestingAggregations}}}

AggregationType = Union[
    AvgAggregation,
    CountAggregation,
    MaxAggregation,
    MinAggregation,
    SumAggregation,
    StatsAggregation,
    TermsAggregation,
    FilterAggregation,
    EntityAggregation,
    DateHistogramAggregation,
]
